from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional

import glm
import moderngl

from moderngl_window.opengl.vao import VAO

from .material import Material

if TYPE_CHECKING:
    from .programs import MeshProgram


class Mesh:
    """Mesh info and geometry"""

    def __init__(
        self,
        name: str,
        vao: Optional[VAO] = None,
        material: Optional[Material] = None,
        attributes: Optional[dict[str, Any]] = None,
        bbox_min: glm.vec3 = glm.vec3(),
        bbox_max: glm.vec3 = glm.vec3(),
    ) -> None:
        """Initialize mesh.

        Args:
            name (str): name of the mesh
        Keyword Args:
            vao (VAO): geometry
            material (Material): material for the mesh
            attributes (dict): Details info about each mesh attribute (dict)
            bbox_min: xyz min values
            bbox_max: xyz max values

        Attributes example::

            {
                "NORMAL": {"name": "in_normal", "components": 3, "type": GL_FLOAT},
                "POSITION": {"name": "in_position", "components": 3, "type": GL_FLOAT}
            }
        """
        self.name = name
        self.vao = vao
        self.material = material
        self.attributes = attributes or {}
        self.bbox_min = bbox_min
        self.bbox_max = bbox_max
        self.mesh_program: Optional["MeshProgram"] = None

    def draw(
        self,
        projection_matrix: glm.mat4,
        model_matrix: glm.mat4,
        camera_matrix: glm.mat4,
        time: float = 0.0,
    ) -> None:
        """Draw the mesh using the assigned mesh program

        Keyword Args:
            projection_matrix (bytes): projection_matrix
            view_matrix (bytes): view_matrix
            camera_matrix (bytes): camera_matrix
        """
        if self.mesh_program is not None:
            self.mesh_program.draw(
                self,
                projection_matrix=projection_matrix,
                model_matrix=model_matrix,
                camera_matrix=camera_matrix,
                time=time,
            )

    def draw_bbox(
        self,
        proj_matrix: glm.mat4,
        model_matrix: glm.mat4,
        cam_matrix: glm.mat4,
        program: moderngl.Program,
        vao: VAO,
    ) -> None:
        """Renders the bounding box for this mesh.

        Args:
            proj_matrix: Projection matrix
            model_matrix: View/model matrix
            cam_matrix: Camera matrix
            program: The moderngl.Program rendering the bounding box
            vao: The vao mesh for the bounding box
        """
        program["m_proj"].write(proj_matrix.to_bytes())
        program["m_model"].write(model_matrix.to_bytes())
        program["m_cam"].write(cam_matrix.to_bytes())
        program["bb_min"].write(self.bbox_min.to_bytes())
        program["bb_max"].write(self.bbox_max.to_bytes())
        vao.render(program)

    def draw_wireframe(
        self, proj_matrix: glm.mat4, model_matrix: glm.mat4, program: moderngl.Program
    ) -> None:
        """Render the mesh as wireframe.

        proj_matrix: Projection matrix
        model_matrix: View/model matrix
        program: The moderngl.Program rendering the wireframe
        """
        assert self.vao is not None, "Can not draw the wireframe, vao is empty"
        program["m_proj"].write(proj_matrix.to_bytes())
        program["m_model"].write(model_matrix.to_bytes())
        self.vao.render(program)

    def add_attribute(self, attr_type: str, name: str, components: int) -> None:
        """
        Add metadata about the mesh
        :param attr_type: POSITION, NORMAL etc
        :param name: The attribute name used in the program
        :param components: Number of floats
        """
        self.attributes[attr_type] = {"name": name, "components": components}

    def calc_global_bbox(
        self, view_matrix: glm.mat4, bbox_min: glm.vec3 | None, bbox_max: glm.vec3 | None
    ) -> tuple[glm.vec3, glm.vec3]:
        """Calculates the global bounding.

        Args:
            view_matrix: View matrix
            bbox_min: xyz min
            bbox_max: xyz max
        Returns:
            bbox_min, bbox_max: Combined bbox
        """
        # Copy and extend to vec4
        bb1 = glm.vec4(self.bbox_min, 1.0)
        bb2 = glm.vec4(self.bbox_max, 1.0)

        # Transform the bbox values
        bmin = view_matrix * bb1
        bmax = view_matrix * bb2

        # If a rotation happened there is an axis change and we have to ensure max-min is positive
        for i in range(3):
            if bmax[i] - bmin[i] < 0:
                bmin[i], bmax[i] = bmax[i], bmin[i]

        if bbox_min is None or bbox_max is None:
            return (glm.vec3(bmin.x, bmin.y, bmin.z), glm.vec3(bmax.x, bmax.y, bmax.z))

        for i in range(3):
            bbox_min[i] = min(bbox_min[i], bmin[i])

        for i in range(3):
            bbox_max[i] = max(bbox_max[i], bmax[i])

        return bbox_min, bbox_max

    def has_normals(self) -> bool:
        """
        Returns:
            bool: Does the mesh have a normals?
        """
        return "NORMAL" in self.attributes

    def has_uvs(self, layer: int = 0) -> bool:
        """
        Returns:
            bool: Does the mesh have texture coordinates?
        """
        return "TEXCOORD_{}".format(layer) in self.attributes



from moderngl_window.geometry.attributes import AttributeNames as AttributeNames
from moderngl_window.geometry.bbox import bbox as bbox
from moderngl_window.geometry.cube import cube as cube
from moderngl_window.geometry.quad import quad_2d as quad_2d
from moderngl_window.geometry.quad import quad_fs as quad_fs
from moderngl_window.geometry.sphere import sphere as sphere

__all__ = [
    "AttributeNames",
    "bbox",
    "cube",
    "quad_2d",
    "quad_fs",
    "sphere",
]



"""
Scene Registry
"""

from moderngl_window.meta import ResourceDescription, SceneDescription
from moderngl_window.resources.base import BaseRegistry
from moderngl_window.scene import Scene


class Scenes(BaseRegistry):
    """Handles scene loading"""

    settings_attr = "SCENE_LOADERS"
    meta: SceneDescription

    def load(self, meta: ResourceDescription) -> Scene:
        """Load a scene with the configured loaders.

        Args:
            meta (:py:class:`~moderngl_window.meta.scene.SceneDescription`):
            The resource description
        Returns:
            :py:class:`~moderngl_window.scene.Scene`: The loaded scene
        """
        scene = super().load(meta)
        assert isinstance(
            scene, Scene
        ), f"{meta} did not load a moderngl_window.scene.Scene object, please correct it."
        return scene


scenes = Scenes()



"""
Example showing the use of #include in shaders.
This can be used to include reusable library functions.

We include a library doing different blend types
and render each quadrant of the screen with different blend types
"""

from pathlib import Path

import moderngl_window as mglw
from moderngl_window import geometry


class ShaderInclude(mglw.WindowConfig):
    title = "Shader Include"
    resource_dir = (Path(__file__) / "../../resources").resolve()
    aspect_ratio = 1

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.program = self.load_program("programs/blend_include.glsl")
        self.texture_0 = self.load_texture_2d("textures/cubemaps/yokohama/negx.jpg")
        self.texture_1 = self.load_texture_2d("textures/cubemaps/yokohama/negz.jpg")
        self.quad_fs = geometry.quad_fs()

    def on_render(self, time, frame_time):
        self.texture_0.use(location=0)
        self.texture_1.use(location=1)
        self.quad_fs.render(self.program)


if __name__ == "__main__":
    ShaderInclude.run()



import subprocess
from typing import Any, Optional

import moderngl

from .base import BaseVideoCapture


class FFmpegCapture(BaseVideoCapture):
    """
    ``FFmpegCapture`` it's an utility class to capture runtime render
    and save it as video.

    Args:


    Example:

    .. code:: python

        import moderngl_window
        from moderngl_window.capture.ffmpeg import FFmpegCapture

        class CaptureTest(modenrgl_window.WindowConfig):

            def __init__(self, **kwargs):
                super().__init__(**kwargs)
                # do other initialization

                # define VideoCapture class
                self.cap = FFmpegCapture(source=self.wnd.fbo)

                # start recording
                self.cap.start_capture(
                    filename="video.mp4",
                    framerate=30
                )

            def render(self, time, frametime):
                # do other render stuff

                # call record function after
                self.cap.save()

            def close(self):
                # if realease func is not called during
                # runtime. make sure to do it on the closing event
                self.cap.release()

    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._ffmpeg: Optional[subprocess.Popen[bytes]] = None

    def _start_func(self) -> bool:
        """
        choose the right pixel format based on the number of components
        and start a ffmpeg pipe with a subprocess.
        """
        pix_fmt = "rgb24"  # 3 component, 1 byte per color -> 24 bit

        # for the framebuffer is easier because i can read 3 component even if
        # the color attachment has less components
        if isinstance(self._source, moderngl.Texture) and self._components == 4:
            pix_fmt = "rgba"  # 4 component , 1 byte per color -> 32 bit

        command = [
            "ffmpeg",
            "-hide_banner",
            "-loglevel",
            "error",
            "-stats",  # less verbose, only stats of recording
            "-y",  # (optional) overwrite output file if it exists
            "-f",
            "rawvideo",
            "-vcodec",
            "rawvideo",
            "-s",
            f"{self._width}x{self._height}",  # size of one frame
            "-pix_fmt",
            pix_fmt,
            "-r",
            f"{self._framerate}",  # frames per second
            "-i",
            "-",  # The imput comes from a pipe
            "-vf",
            "vflip",
            "-an",  # Tells FFMPEG not to expect any audio
            self._filename,
        ]

        # ffmpeg binary need to be on the PATH.
        try:
            self._ffmpeg = subprocess.Popen(command, stdin=subprocess.PIPE, bufsize=0)
        except FileNotFoundError:
            print("ffmpeg command not found. Be sure to add it to PATH")
            return False

        return True

    def _release_func(self) -> None:
        """
        Safely release the capture
        """
        if (self._ffmpeg is None) or (self._ffmpeg.stdin is None):
            return
        self._ffmpeg.stdin.close()
        _ = self._ffmpeg.wait()

    def _dump_frame(self, frame: Any) -> None:
        """
        write the frame data in to the ffmpeg pipe
        """
        if (self._ffmpeg is None) or (self._ffmpeg.stdin is None):
            return
        self._ffmpeg.stdin.write(frame)



from pathlib import Path

import glm
import moderngl
from base import CameraWindow

import moderngl_window
from moderngl_window import geometry


class CubeSimple(CameraWindow):
    title = "Plain Cube"
    resource_dir = (Path(__file__).parent / "resources").resolve()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.wnd.mouse_exclusivity = True
        self.cube = geometry.cube(size=(2, 2, 2))
        self.prog = self.load_program("programs/cube_simple.glsl")
        self.prog["color"].value = 1.0, 1.0, 1.0, 1.0

    def on_render(self, time: float, frametime: float):
        self.ctx.enable_only(moderngl.CULL_FACE | moderngl.DEPTH_TEST)

        rotation = glm.mat4(glm.quat(glm.vec3(time, time, time)))
        translation = glm.translate(glm.vec3(0.0, 0.0, -3.5))
        modelview = translation @ rotation

        self.prog["m_proj"].write(self.camera.projection.matrix)
        self.prog["m_model"].write(modelview)
        self.prog["m_camera"].write(self.camera.matrix)

        self.cube.render(self.prog)


if __name__ == "__main__":
    moderngl_window.run_window_config(CubeSimple)
